#include <Color4.hpp>
#include <format>
#include <ColorFormats.hpp>

#if !defined(WIN32)
// For some reason this causes crash with exit code: -1073741819 on winblows.
std::vector<Color4> list;
#endif

std::string toEscapeCode(Color4 c, bool bg){
    if (bg)
        return std::format("\033[48;2;{};{};{}m", c.r, c.g, c.b);

    return std::format("\033[38;2;{};{};{}m", c.r, c.g, c.b);
}

void print_builtin_color_list() {
#if !defined(WIN32)
    int i = 0;
    for (const Color4& color : list) {
        std::cout << toEscapeCode(color) <<  i << ": " << "Hello, world!" << std::endl;
        i++;
    }
#endif
}

Color4::Color4(const RGB &rgb, u8 alpha): Color4(rgb.r, rgb.g, rgb.b, alpha) {}

Color4::Color4(const RGBA &rgba): Color4(rgba.r, rgba.g, rgba.b, rgba.a) {}

Color4::Color4(const RGBf &rgb, float alpha): Color4(rgb.r * 255, rgb.g * 255, rgb.b * 255, alpha * 255) { }

Color4::Color4(const RGBAf &rgba): Color4(rgba.r * 255, rgba.g * 255, rgba.b * 255, rgba.a * 255) { }

Color4::Color4(const HSV &hsv, float alpha) {
    float h = hsv.h;
    float s = hsv.s;
    float v = hsv.v;
    float M = 255*hsv.v;
    float m = M*(1-hsv.s);

    float z = (M-m)*(1 - std::abs( std::fmod(hsv.h/60.f, 2) - 1));

    if (0 <= h <= 60) {
        r = M;
        g = z + m;
        b = m;
    }
    if (60 <= h < 120) {
        r = z + m;
        g = M;
        b = m;
    }
    if (120 <= h < 180) {
        r = m;
        g = M;
        b = z + m;
    }

    if (180 <= h < 240) {
        r = m;
        g = z + m;
        b = M;
    }

    if (240 <= h < 300) {
        r = z + m;
        g = m;
        b = M;
    }

    if (300 <= h < 360) {
        r = M;
        g = m;
        b = z + m;
    }

    a = alpha;
}

Color4::Color4(const Color3 &color3, u8 alpha) {r = color3.r; g = color3.g; b = color3.b; a = alpha;}

Color4::Color4(u8 red, u8 green, u8 blue, u8 alpha): r(red), g(green), b(blue), a(alpha) {
#if !defined(WIN32)
    list.push_back(*this);
#endif
}

Color4 Color4::FromColor3(const Color3 &color3, u8 alpha) {return Color4(color3, alpha);}

Color4 Color4::FromHex(const std::string &hexCode, u8 alpha) {
    // TODO: Support hex codes without the #
    // TODO: Support 9-character hex codes (with alpha), but raise a warning that suggests using FromHexA explicitly instead.
    if (hexCode.length() == 7) {
        u8 r, g, b;
        std::sscanf(hexCode.c_str(), "#%02x%02x%02x", &r, &g, &b);
        return {r, g, b, alpha};
    }

    if (hexCode.length() == 9) {
        u8 r, g, b, a;
        std::sscanf(hexCode.c_str(), "#%02x%02x%02x%02x", &r, &g, &b, &a);
        return {r, g, b, a};
    }

    throw std::invalid_argument("Invalid hex code");
}

Color4 Color4::FromHexA(const std::string &hexACode) {
    if (hexACode.length() == 9) {
        u8 r, g, b, a;
        std::sscanf(hexACode.c_str(), "#%02x%02x%02x%02x", &r, &g, &b, &a);
        return {r, g, b, a};
    }

    throw std::invalid_argument("Invalid hex code");
}

u8 Color4::RedChannel() const { return r;}

u8 Color4::GreenChannel() const { return g;}

u8 Color4::BlueChannel() const {return b;}

u8 Color4::AlphaChannel() const {return a;}

u8 Color4::R() const { return AlphaChannel(); }

u8 Color4::G() const { return GreenChannel(); }

u8 Color4::B() const { return BlueChannel();  }

u8 Color4::A() const { return AlphaChannel(); }

float Color4::RedChannelNormalized() const {return static_cast<float>(r/255.f); }

float Color4::GreenChannelNormalized() const {return static_cast<float>(g/255.f); }

float Color4::BlueChannelNormalized() const {return static_cast<float>(b/255.f); }

float Color4::AlphaChannelNormalized() const {return static_cast<float>(a/255.f); }

float Color4::RN() const { return RedChannelNormalized(); }

float Color4::GN() const { return GreenChannelNormalized(); }

float Color4::BN() const { return BlueChannelNormalized(); }

float Color4::AN() const { return AlphaChannelNormalized(); }


Color4 Color4::FromNormalized(float red, float green, float blue, float alpha) {
    return {
            static_cast<u8>(red * 255),
            static_cast<u8>(green * 255),
            static_cast<u8>(blue * 255),
            static_cast<u8>(alpha * 255)
    };

}

Color4 Color4::Lerp(const Color4 &rhs, float t) const {

    /// Performs the interpolation in normalized color space (0-1) and converts it back.



    return FromNormalized(
            std::lerp(RN(), rhs.RN(), t),
            std::lerp(GN(), rhs.GN(), t),
            std::lerp(BN(), rhs.BN(), t),
            std::lerp(AN(), rhs.AN(), t)
    );
}

Color4 Color4::Lerp2(const Color4& rhs, float t) const {
    return Color4(
           std::lerp(R(), rhs.R(), t),
           std::lerp(G(), rhs.G(), t),
           std::lerp(B(), rhs.B(), t),
           std::lerp(A(), rhs.A(), t)
   );
}

Color4 Color4::Lerp2(const Color4 &lhs, const Color4 &rhs, float t) {
    return lhs.Lerp2(rhs, t);
}

Color4 Color4::LerpByHSVA(const Color4 &rhs, float t) const {
    HSVA a = this->ToHSVA();
    HSVA b = rhs.ToHSVA();

    // Hue interpolation
    float h;
    float d = b.h - a.h;

    if (a.h > b.h) {
        auto h3 = b.h; // b.h2?
        b.h = a.h;
        a.h = h3;

        d = -d;
        t = 1 - t;
    }

    if (d > 0.5f) { // 180 deg
        a.h = a.h + 1; // 360 deg
        h = std::fmod(a.h + t * (b.h - a.h), 1.f);
    }

    if (d <= 0.5f) { // 180 deg
        h = a.h + t * d;
    }

    return FromHSV(
        h,
        std::lerp(a.s, b.s, t),
        std::lerp(a.v, b.v, t),
        std::lerp(a.a, b.a, t));
}

Color4 Color4::Lerp(const Color4 &lhs, const Color4 &rhs, float t) {
    return lhs.Lerp(rhs, t);
}

u8 *Color4::ptr() { return (&r);}

const u8 *Color4::ptr() const { return (&r);}

HSV Color4::ToHSV() const {
    float rn = RN();
    float gn = GN();
    float bn = BN();
    auto Cmax = std::max(std::max(rn, gn), bn); // TODO: Replace with initializer list.
    auto Cmin = std::min(std::max(rn, gn), bn); // TODO: Replace with initializer list.

    float delta = Cmax - Cmin;

    float hue = 0.f;

    if (Cmax == rn) hue = 60.f * std::fmod((gn - bn) / delta,  6.0);
    if (Cmax == gn) hue = 60.f * ((gn - rn) / delta) + 2;
    if (Cmax == bn) hue = 60.f * ((rn - gn)/ delta) + 4;

    float saturation = 0.f;

    if (Cmax == 0) saturation = 0.f;
    else           saturation = delta / Cmax;

    float value = Cmax;

    return {hue, saturation, value};
}

HSVA Color4::ToHSVA() const {
    auto hsv = ToHSV();
    return {hsv.h, hsv.s, hsv.v, AN()};
}

HSL Color4::ToHSL() const
{
    float rn = RN();
    float gn = GN();
    float bn = BN();

    float max = std::max(std::max(rn, gn), bn);
    float min = std::min(std::min(rn, gn), bn);

    float delta = max - min;

    float L = (min + max) / 2.f;
    float S = 0;

    if (max == min)
        S = 0;
    if (L > 0.5f)
        S = (delta) * (2.f - max - min);
    if (L <= 0.5f)
        S = (max - min) / (max + min);

    float H = 0;

    if (rn == max)
        H = (gn - bn) / (delta);
    if (gn == max)
        H = 2.f + (bn-rn) / (max-min);
    if (bn == max)
        H = 4.f + (rn-gn) / (max-min);

    H *= 60;

    if (H < 0.f)
        H += 360.f;

    return {H,S,L};
}

HSLA Color4::ToHSLA() const
{
    HSL hsl = ToHSL();
    return {hsl.h, hsl.s, hsl.l, AN()};
}

LCH Color4::ToLCH() const {
    // TODO: Implement
    return {};
}
